{
=============================================================================
| THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  |
| AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    |
| IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   |
| ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR  |
| ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL       |
| DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR   |
| SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER   |
| CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT           |
| LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY    |
| OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  |
| DAMAGE.                                                                      |
=============================================================================

  Author: D.J.C. Blijker
  Date:  2-4-2003
  Versie: 1.0
  e-mail: c.blijker@planet.nl

  Class: THxDataExchange

=============================================================================
This unit will work in Delphi as well in Kylix!!!
=============================================================================
Main target of this unit is to supply a consistent and easy to use means to
exchange data over a network.

usage:
-Instantiate an object of THxDataExchange
-Call Readxxx() to read elements out of datastream packed by a THxDataExchange
 component
-To create a 'Message'
  -Initiate a new message with a call to StartCreatingMessage;
  -Call Addxxxx() as many time as necessary
  -Call GetCreatedMessage( var pMess: pchar; var iSize: integer );
    After this call pMess will point to the newly created message and
    iSize will hold the size of the message.
    ** Don't Dispose/Free ptr!! It is maintained by this class.

Below you'll find some examples.
 ( 'deDataExchange' is an instantion of the class THxDataExchange, for example
   created in the constructor of class xxx )

procedure xxx.SendRecord( iType: integer; sMess: string; bVal: boolean;
                           fValue: double );
var
  ptr: pchar;
  iSize: integer;
begin
// Always start a new message with a call of StartCreatingMessage..
  deDataExchange.StartCreatingMessage;

// Add all parameters..
  deDataExchange.AddInteger( iType );
  deDataExchange.AddString( sMess );
  deDataExchange.AddBoolean( bVal );
  deDataExchange.AddDouble( fValue );

// Get the newly created message..
  deDataExchange.GetCreatedMessage( ptr, iSize );

// Send the message over an 'ipc-connection'
  ipcClient.SendData( ptr, iSize )
end;

/////////////////////////////////////////////////////////////////////////////
Below an example routine which extracts all parameters from the previous
example message received by an 'ipc-server'

procedure yyy.HandleIpcData( ptr: pointer; iSize, iDescr: integer );
var
  iType: integer;
  sMess: string;
  fValue: double
  bVal: boolean;
begin
// Just extract all the parameters in the right order from the stream..

  ptr := deDataExchange.ReadInteger( ptr, iType [, iSize] );
  ptr := deDataExchange.ReadString( ptr, sMess [, iSize] );
  ptr := deDataExchange.ReadBool( ptr, bVal [, iSize] );
  deDataExchange.ReadDouble( ptr, fValue [, iSize] );

  DoSomething( iType, sMess, bVal, fValue );
end;
==============================================================================
History:

  version: 1.0  Initial public release.
==============================================================================

}
unit HxDataExchange;

interface

uses
{$IFDEF LINUX}
  libc,
{$ELSE}
  windows,
{$ENDIF}
  SysUtils;

const
  INVALID_NUMBER = MAXINT;
  Version = '1.0';
  EOF_Marker = 'EOF_070859_EOF'; // Unieke marker...

type
  THxDataExchange = class
  private
    ptrMess: pchar;
    ptrOffset: pchar;
    iMessSize: integer;
    iAllocSize: integer;
    iChunkSize: integer;
    cOrgDesSepChar: char;
    function ArrayToStr(aArray: array of char): string;
    procedure StrLenCopy(aDest, aSource: pchar; iMaxLength: integer);
    procedure SetChunkSize(const Value: integer);
    procedure CheckMemory;
  public
    constructor Create;
    destructor Destroy; override;
// Change the default 'ChunckSize' only if you plan to create real large
// messages.. The default is sufficient for almost all cases..
    property  ChunkSize: integer read iChunkSize write SetChunkSize;
    procedure StartCreatingMessage;
    procedure AddBool( bVal: Boolean );
    procedure AddString( sVal: string );
    procedure AddArray( aVal: array of char );
    procedure AddInteger( iVal: integer );
// Handle AddPtr and AddData with care because no checks can be made and data
// will be transmitted in its *raw* format.
    procedure AddPtr( ptr: pchar; iSize: integer );
    procedure AddData( pVariable: pointer; iSize: integer );
    procedure AddCurrency( fValue: currency );
    procedure AddDouble( fValue: extended );
    procedure GetCreatedMessage( var ptr: pchar; var iSize: integer );
// The pointer returnd by GetCopyCreatedMessage will *NOT* be maintained
// by this class. So you'll have to dispose it your self once you're done
// with it.
    procedure GetCopyCreatedMessage( var ptr: pchar; var iSize: integer );
    function  ReadBool( ptr: pchar; var bVal: boolean ): pchar; overload;
    function  ReadBool( ptr: pchar; var bVal: boolean;
                        var iSize: integer ): pchar; overload;
    function  ReadInteger( ptr: pchar; var iVal: integer ): pchar; overload;
    function  ReadInteger( ptr: pchar;
                           var iVal, iSize: integer ): pchar; overload;
    function  ReadString( ptr: pchar; var sRes: string ): pchar; overload;
    function  ReadString( ptr: pchar; var sRes: string;
                          var iSize: integer ): pchar; overload;
    function  ReadCurrency( ptr: pchar; var fValue: currency ): pchar; overload;
    function  ReadCurrency( ptr: pchar; var fValue: currency;
                            var iSize: integer ): pchar; overload;
    function  ReadDouble( ptr: pchar; var fValue: extended ): pchar; overload;
    function  ReadDouble( ptr: pchar; var fValue: extended;
                          var iSize: integer ): pchar; overload;
    function  ReadData( ptr: pchar;
                        pVariable: pointer;
                        iSizeOfVar: integer ): pchar; overload;
    function  ReadData( ptr: pchar;
                        pVariable: pointer;
                        iSizeOfVar: integer;
                        var iSize: integer ): pchar; overload;
    function  PeekInteger( ptr: pchar ): integer;
    procedure MarkEof;
    function  Eof( ptr: pchar ): boolean;
 end;

implementation

constructor THxDataExchange.Create;
begin
  ptrMess := nil;
  ptrOffset := nil;
  iMessSize := 0;
  iAllocSize := 0;
  iChunkSize := 512;
  cOrgDesSepChar := Decimalseparator;
end;

destructor THxDataExchange.Destroy;
begin
  if ptrMess <> nil then
    ReAllocMem( ptrMess, 0 );

  inherited;
end;

procedure THxDataExchange.StrLenCopy( aDest, aSource: pchar;
                                      iMaxLength: integer );
var
  l: integer;
  bZerro: boolean;
begin
  l := Length( aSource );
  if l >= iMaxLength then
  begin
    l := iMaxLength;
    bZerro := False;
  end
  else
    bZerro := True;
{$IFDEF LINUX}
  memcpy( aDest, aSource, l );
{$ELSE}
  CopyMemory( aDest, aSource, l );
{$ENDIF}

  if bZerro then
    aDest[l] := #0;
end;

function THxDataExchange.ArrayToStr( aArray: array of char ): string;
begin
  SetLength( result, Sizeof( aArray ));
  StrLenCopy( pchar( result ), aArray, Sizeof( aArray ));
  SetLength( result, StrLen( pchar( result )));
end;

procedure THxDataExchange.CheckMemory;
begin
  if iMessSize > iAllocSize then
  begin
    if iMessSize > iAllocSize + iChunkSize then
      iAllocSize := iMessSize
    else
      Inc( iAllocSize, iChunkSize );

    ReAllocMem( ptrMess, iAllocSize );
  end;
end;

procedure THxDataExchange.StartCreatingMessage;
begin
  ReAllocMem( ptrMess, 0 );
  iMessSize := 0;
  iAllocSize := 0;
  ptrOffset := nil;
end;

procedure THxDataExchange.GetCreatedMessage( var ptr: pchar;
                                              var iSize: integer );
begin
  ptr := ptrMess;
  iSize := iMessSize;
end;

procedure THxDataExchange.GetCopyCreatedMessage( var ptr: pchar;
                                                  var iSize: integer);
begin
  ptr := nil;
  ReAllocMem( ptr, iMessSize );
{$IFDEF LINUX}
  memcpy( ptr, ptrMess, iMessSize );
{$ELSE}
  CopyMemory( ptr, ptrMess, iMessSize );
{$ENDIF}

  iSize := iMessSize;
end;


procedure THxDataExchange.SetChunkSize(const Value: integer);
begin
  if Value > 0 then
    iChunkSize := Value;
end;

function THxDataExchange.ReadString( ptr: pchar; var sRes: string ): pchar;
var
  iDummy: integer;
begin
  result := ReadString( ptr, sRes, iDummy );
end;

function THxDataExchange.ReadString( ptr: pchar; var sRes: string;
                                     var iSize: integer ): pchar;
begin
  sRes := Strpas( ptr );

  Inc( ptr, Length( sRes )+1 );

  Dec( iSize, Length( sRes )+1 );

  result := ptr;
end;

function THxDataExchange.ReadBool( ptr: pchar; var bVal: Boolean ): pchar;
var
  iDummy: integer;
begin
  result := ReadBool( ptr, bVal, iDummy );
end;

function THxDataExchange.ReadBool( ptr: pchar; var bVal: Boolean;
                                   var iSize: integer ): pchar;
var
  sVal: string;
begin
  result := ReadString( ptr, sVal );
  bVal := ( sVal = '1' );
  Dec( iSize, Length( sVal )+1 );
end;

function THxDataExchange.ReadInteger( ptr: pchar; var iVal: integer ): pchar;
var
  iDummy: integer;
begin
  result := ReadInteger( ptr, iVal, iDummy );
end;

function THxDataExchange.ReadInteger( ptr: pchar;
                                      var iVal, iSize: integer ): pchar;
var
  sVal: string;
begin
  result := ReadString( ptr, sVal );
  Dec( iSize, Length( sVal )+1 );
  try
    iVal := StrToInt( sVal );
  except
    iVal := INVALID_NUMBER;
  end;
end;

function THxDataExchange.ReadCurrency( ptr: pchar;
                                       var fValue: currency ): pchar;
var
  iDummy: integer;
begin
  result := ReadCurrency( ptr, fValue, iDummy );
end;

function THxDataExchange.ReadCurrency( ptr: pchar; var fValue: currency;
                                       var iSize: integer ): pchar;
var
  sVal: string;
begin
  result := ReadString( ptr, sVal );
  Dec( iSize, Length( sVal )+1 );
  try
    try
      DecimalSeparator := ',';
      fValue := StrToCurr( sVal );
    finally
      DecimalSeparator := cOrgDesSepChar;
    end;
  except
    fValue := INVALID_NUMBER;
  end;
end;

function THxDataExchange.ReadDouble( ptr: pchar;
                                     var fValue: extended ): pchar;
var
  iDummy: integer;
begin
  result := ReadDouble( ptr, fValue, iDummy );
end;

function THxDataExchange.ReadDouble( ptr: pchar; var fValue: extended ;
                                     var iSize: integer ): pchar;
var
  sVal: string;
begin
  result := ReadString( ptr, sVal );
  Dec( iSize, Length( sVal )+1 );
  try
    try
      DecimalSeparator := ',';
      fValue := StrToFloat( sVal );
    finally
      DecimalSeparator := cOrgDesSepChar;
    end;
  except
    fValue := INVALID_NUMBER;
  end;
end;

function THxDataExchange.ReadData( ptr: pchar;
                                   pVariable: pointer;
                                   iSizeOfVar: integer ): pchar;
var
  iDummy: integer;
begin
  result := ReadData( ptr, pVariable, iSizeOfVar, iDummy );
end;

function THxDataExchange.ReadData( ptr: pchar;
                                   pVariable: pointer;
                                   iSizeOfVar: integer;
                                   var iSize: integer ): pchar;
begin
{$IFDEF LINUX}
  memcpy( pVariable, ptr, iSizeOfVar );
{$ELSE}
  CopyMemory( pVariable, ptr, iSizeOfVar );
{$ENDIF}
  Inc( ptr, iSizeOfVar );
  Dec( iSize, iSizeOfVar );
  result := ptr;
end;


procedure THxDataExchange.AddInteger( iVal: integer );
var
  sVal: string;
begin
  sVal := IntToStr( iVal );
  AddString( sVal );
end;

procedure THxDataExchange.AddBool( bVal: Boolean );
var
  sVal: string;
begin
  if bVal then
    sVal:= '1'
  else
    sVal:= '0';
  AddString( sVal );
end;

procedure THxDataExchange.AddString( sVal: string );
var
  iSize: integer;
begin
  iSize := Length( sVal )+ 1;
  Inc( iMessSize, iSize );

  CheckMemory;

  ptrOffset := ptrMess;
  Inc( ptrOffset, iMessSize );
  Dec( ptrOffset, iSize );

  StrCopy( ptrOffset, pchar( sVal ));
end;

procedure THxDataExchange.AddCurrency( fValue: currency );
var
  sVal: string;
begin
  try
    DecimalSeparator := ',';
    sVal := CurrToStr( fValue );
    AddString( sVal );
  finally
    DecimalSeparator := cOrgDesSepChar;
  end;
end;

procedure THxDataExchange.AddDouble( fValue: extended );
var
  sVal: string;
begin
  try
    DecimalSeparator := ',';
    sVal := FloatToStr( fValue );
    AddString( sVal );
  finally
    DecimalSeparator := cOrgDesSepChar;
  end;
end;

procedure THxDataExchange.AddPtr( ptr: pchar; iSize: integer );
begin
  Inc( iMessSize, iSize );

  CheckMemory;

  ptrOffset := ptrMess;
  Inc( ptrOffset, iMessSize );
  Dec( ptrOffset, iSize );
{$IFDEF LINUX}
  memcpy( ptrOffset, ptr, iSize );
{$ELSE}
  CopyMemory( ptrOffset, ptr, iSize );
{$ENDIF}
end;

procedure THxDataExchange.AddData( pVariable: pointer; iSize: integer );
begin
  AddPtr( pVariable, iSize );
end;

procedure THxDataExchange.AddArray( aVal: array of char );
begin
  AddString( ArrayToStr( aVal ));
end;

procedure THxDataExchange.MarkEof;
begin
  AddString( EOF_Marker );
end;

// Read the next string, check whether this an EOF_marker but leave the
// pointer unchanged..
function THxDataExchange.Eof( ptr: pchar ): boolean;
var
  s: string;
begin
  ReadString( ptr, s );
  result := s = EOF_Marker;
end;

// Read the next integer value but leave ptr unchanged.
function THxDataExchange.PeekInteger( ptr: pchar ): integer;
var
  iVal: integer;
begin
  ReadInteger( ptr, iVal );
  result := iVal;
end;

end.
